"""ricloud API Client.

Usage:
    ricloud <account> [--password=<password>] [--timeout=<timeout>]

Options:
    -h --help               Show this screen.
    --password=<password>   The password for this account.
    --timeout=<timeout>     How long should we wait for tasks to complete, in seconds. (default: 600s or 10 minutes)

"""
from __future__ import unicode_literals

from docopt import docopt

from .ricloud import RiCloud
from .conf import DEFAULT_LOG
from .utils import LogFile, select_service, select_samples


def _parse_input_arguments(arguments):
    account = arguments['<account>']
    password = arguments.get('--password', None)
    timeout = int(arguments.get('--timeout') or 600)

    return {
        'account': account,
        'password': password,
        'timeout': timeout
    }


def main():
    # Get input arguments from command line
    arguments = docopt(__doc__)

    # Format the input arguments for the application
    application_payload = _parse_input_arguments(arguments)

    # Main RiCloud object
    ricloud = RiCloud(log=LogFile(DEFAULT_LOG))

    # The RiCloud client retrieves a list of available services from the API.
    service_name = select_service(ricloud)

    # Each service has a client which contains all the available tasks.
    # Sample applications for these clients can be found in the `samples` folder.
    Application = select_samples(ricloud, service_name, application_payload)

    # Execute the application's main function.
    Application.run()


if __name__ == '__main__':
    main()



import json
import os

from clint.textui import puts, colored
from ..clients.icloud import iCloudClient

from ..conf import OUTPUT_DIR

SAMPLE_ACCOUNT = "johnapplessed@apple.com"
SAMPLE_PASSWORD = "johnapple"
SAMPLE_DEVICE_ID = 'b5cc9d671f96f4443d522ac15ca8bc9f66f18cf5'

#
# Task Submission and Waiting for results:
#   By calling ricloud.api.perform_task, a task is submitted to the push api.
#   The perform_task method returns the task object.
#
#   The result of the task comes from a separate thread created upon the
#   intialization of the ricloud client. This thread is connected to the client's
#   stream and saves the task result in memory. This result can be recovered by
#   calling the task.wait_for_task_result() method.
#
#   If the task's result is already in memory, this method returns immediately.
#   If not, the method blocks until the push api returns the task result.
#   If no result comes before the the timeout, the method raises an Exception
#


class SampleICloudNativeApplication(object):
    display_name = 'Sample iCloud Native Application'
    client_name = iCloudClient
    service = 'iCloud'

    def __init__(self, ricloud_client, payload):
        self.ricloud = ricloud_client
        self.account = SAMPLE_ACCOUNT
        self.password = SAMPLE_PASSWORD
        self.device_id = SAMPLE_DEVICE_ID
        self.data_to_fetch = ','.join(self.get_available_data())
        self.timeout = payload.get('timeout')

    def run(self):
        # Register the account for the iCloud service
        self.ricloud.api.register_account(self.account, self.service)

        # Log in to the sample account
        task = self.submit_login_task()
        # Wait for the task's result
        login_result = task.wait_for_result()
        self.handle_login_response(login_result)

        # Submit a task to get all available data
        task = self.submit_fetch_data_task()
        # Wait for the task's result
        feed_result = task.wait_for_result()

        # Write feed to disk
        self.write_feed_to_disk(feed_result, task.uuid)

    def submit_login_task(self):

        task = self.ricloud.api.perform_task(
            service=self.service,
            task_name='log-in',
            account=self.account,
            payload={
                'password': self.password,
            },
        )
        return task

    def submit_fetch_data_task(self):

        # Fetch data feed
        task = self.ricloud.api.perform_task(
            service=self.service,
            task_name='fetch-data',
            account=self.account,
            payload={
                'data': self.data_to_fetch,
                'device': self.device_id
            },
        )
        return task

    @staticmethod
    def write_feed_to_disk(feed, uuid):
        filename = os.path.join(OUTPUT_DIR, '%s.json' % uuid)
        with open(filename, 'w') as out:
            json.dump(feed, out, indent=2)
        puts(colored.green(
            '\nData Fetch Successful. Output file: %s.json' % uuid)
        )
        puts(colored.green('All tasks completed.'))

    def get_available_data(self):
        if 'permissions' in self.ricloud.api.services['iCloud']['Fetch Data']:
            return self.ricloud.api.services[
                'iCloud']['Fetch Data']['permissions']['data']
        return ''

    @staticmethod
    def handle_login_response(login_result):
        is_login_success = ('error' not in login_result)
        if not is_login_success:
            raise Exception("Login to sample account failed")



"""Primary API object. This handles communication with the ASApi enpoints."""

from __future__ import absolute_import

import time
import requests

from .conf import settings


class Api(object):
    """Primary object that pushes requests into a distinct stream thread."""

    def __init__(self):
        self.host = settings.get('hosts', 'api_host')
        self.token = settings.get('auth', 'token')

        self.account_info_endpoint = '%s%s' % (self.host, settings.get('endpoints', 'account_information'))
        self.register_account_endpoint = '%s%s' % (self.host, settings.get('endpoints', 'register_account'))
        self.task_status_endpoint = '%s%s' % (self.host, settings.get('endpoints', 'task_status'))
        self.results_consumed_endpoint = '%s%s' % (self.host, settings.get('endpoints', 'result_consumed'))

        self._pending_tasks = {}
        self.services = {}

    @property
    def pending_tasks(self):
        return self._pending_tasks

    @property
    def token_header(self):
        return {
            'Authorization': 'Token %s' % self.token,
        }

    def _get_info(self):
        """Fetch account information from ASApi host."""
        return self._perform_get_request(self.account_info_endpoint, headers=self.token_header)

    @staticmethod
    def _parse_endpoint(endpoint):
        """Expect endpoint to be dictionary containing `protocol`, `host` and `uri` keys."""
        return "{protocol}://{host}{uri}".format(**endpoint)

    def _set_endpoints(self, info):
        self.stream_endpoints = info['stream_endpoints']

        submission = info['task_submission_endpoint']
        self.submit_endpoint = self._parse_endpoint(submission)

    def _set_allowed_services_and_actions(self, services):
        """Expect services to be a list of service dictionaries, each with `name` and `actions` keys."""
        for service in services:
            self.services[service['name']] = {}

            for action in service['actions']:
                name = action.pop('name')
                self.services[service['name']][name] = action

    def setup(self):
        info = self._get_info()
        self._set_endpoints(info)
        self.retrieval_protocol = info['retrieval_protocol']
        self._set_allowed_services_and_actions(info['services'])

    def allowed_services(self):
        return self.services.keys()

    def allowed_actions(self, service_name):
        return self.services[service_name].keys()

    def register_account(self, username, service):
        """Register an account against a service.
        The account that we're querying must be referenced during any
        future task requests - so we know which account to link the task
        too.
        """
        data = {
            'service': service,
            'username': username,
        }

        return self._perform_post_request(self.register_account_endpoint, data, self.token_header)

    def perform_task(self, service, task_name, account, payload, callback=None):
        """Submit a task to the API.
        The task is executed asyncronously, and a Task object is returned.
        """
        data = {
            'service': service,
            'action': task_name,
            'account': account,
        }
        data.update(payload)

        response = self._perform_post_request(self.submit_endpoint, data, self.token_header)

        task = Task(uuid=response['task_id'], callback=callback)
        self._pending_tasks[task.uuid] = task

        return task

    def task_status(self, task_id):
        """Find the status of a task."""
        data = {
            'task_ids': task_id,
        }
        return self._perform_post_request(self.task_status_endpoint, data, self.token_header)

    def result_consumed(self, task_id):
        """Report the result as successfully consumed."""
        data = {
            'task_ids': task_id,
        }
        return self._perform_post_request(self.results_consumed_endpoint, data, self.token_header)

    def wait_for_results(self):
        while self._pending_tasks:
            time.sleep(0.1)

    def set_task_result(self, task_id, result):
        if task_id not in self._pending_tasks:
            time.sleep(1)
        try:
            self._pending_tasks[task_id].result = result
            self._pending_tasks.pop(task_id)
            self.result_consumed(task_id)
        except KeyError:
            pass

    @staticmethod
    def _parse_response(response, post_request=False):
        if not response.ok:
            response.raise_for_status()

        data = response.json()

        if post_request and not data['success']:
            raise Exception('Push Api Error: [%s]' % data['error'])

        return data

    @staticmethod
    def _perform_get_request(url, headers=None):
        response = requests.get(
            url,
            headers=headers
        )

        return Api._parse_response(response)

    @staticmethod
    def _perform_post_request(url, data, headers=None):
        response = requests.post(
            url,
            data=data,
            headers=headers
        )

        return Api._parse_response(response, post_request=True)


class Task(object):
    """Simple object to encapsulate the result of a task.

    Can also perform checks against the server to determine its status.
    """
    def __init__(self, uuid, callback=None, object_store=False):
        self.uuid = uuid
        self._result = None
        self._resolved = False
        self.callback = callback
        self.object_store = object_store

        self.timer = time.time()

    @property
    def result(self):
        if not self._resolved:
            raise Exception('Task has not completed, no result available.')
        return self._result

    @result.setter
    def result(self, value):
        self._result = value
        self._resolved = True

        start_time = self.timer
        self.timer = time.time() - start_time

        if self.callback:
            self.callback(self)

    def wait_for_result(self, timeout=None):
        start = time.time()

        while not self.has_resolved:
            time.sleep(0.1)
            if timeout and start + timeout < time.time():
                raise Exception('Task failed to complete within the timeout limit.')

        return self._result

    @property
    def has_resolved(self):
        return self._resolved



from cStringIO import StringIO
import tempfile
import os


class StreamError(Exception):

    def __init__(self, message):
        self.message = message


class HandlerBase(object):
    """Interface for all handler objects."""

    TYPE = ''

    def handle(self, header, payload):
        raise NotImplementedError()


class SystemHandler(HandlerBase):
    TYPE = 'system'

    def handle(self, header, payload):
        print '\n-*- System Message -*-\n'
        print 'Header:  ', header
        print 'Paylaod: ', payload
        print '\n'


class ChunkedDataHandler(HandlerBase):
    """Messages can come in chunks, piece them together here."""
    tmp_handles = {}

    def handle(self, header, payload):
        if header.get('total_chunks', 1) > 1:
            # Open our tmp file and write to the location
            stream = self._get_stream(header['task_id'])
            stream.seek(header['chunk_size'] * (header['chunk'] - 1))
            stream.write(payload)
            if header['chunk'] == header['total_chunks']:
                # Message should be complete
                stream.seek(0)
                self.on_complete_message(header, stream)
                stream.close()
                # Remove the stream handle
                del self.tmp_handles[header['task_id']]

        else:
            self.on_complete_message(header, StringIO(payload))

    def _get_stream(self, id):
        if id not in self.tmp_handles:
            self.tmp_handles[id] = tempfile.TemporaryFile()
        return self.tmp_handles[id]

    def on_complete_message(self, header, payload):
        print '-*- message completed -*-'
        print header
        print payload.read()


class LogInHandler(ChunkedDataHandler):
    TYPE = 'log-in'


class DataHandler(ChunkedDataHandler):
    TYPE = 'data'


class FileDownloadHandler(ChunkedDataHandler):
    TYPE = 'download-file'

    def on_complete_message(self, header, stream):
        # Copy the contents from the tmp file to a new location
        filename = header.get('filename', header['task_id'])
        filename = os.path.join('files', filename)
        with open(filename, 'wb') as out:
            while True:
                data = stream.read(1024)
                if not data:
                    break
                out.write(data)


class RiCloudHandler(ChunkedDataHandler):
    TYPE = '__ALL__'

    def __init__(self, api=None, callback=None):
        self.api = api
        self.callback = callback

    def on_complete_message(self, header, stream):
        if header['type'] == 'system':
            message = stream.read()
            if "error" in message:
                raise StreamError(message)
        else:
            if self.callback:
                self.callback(header['task_id'], stream.read())
            elif self.api:
                self.api.set_task_result(header['task_id'], stream.read())
            else:
                raise
